from __future__ import annotations

import sys
from abc import ABC, ABCMeta
from collections.abc import Callable
from typing import get_type_hints

__all__ = ["AnnotatedABC", "AnnotatedABCMeta"]


def _normalize_annotations(fn: Callable[..., object]) -> dict[str, object]:
    module_name = getattr(fn, "__module__", None)
    module = sys.modules.get(module_name) if module_name is not None else None
    globalns = vars(module) if module is not None else {}
    return get_type_hints(fn, globalns=globalns, localns=globalns)


def _has_fn(
    impl: type,
    abc: type,
    name: str,
    expected: dict[str, object],
) -> bool:
    skip = set(abc.__mro__)
    for base in impl.__mro__:
        if base in skip:
            continue
        for key, value in base.__dict__.items():
            if key == name and _normalize_annotations(value) == expected:
                return True
    return False


class AnnotatedABCMeta(ABCMeta):
    def register(cls, subclass: type) -> type:
        cls._validate_implementation(subclass)
        return super().register(subclass)

    def _validate_implementation(cls, impl: type) -> None:
        for name, expected in cls._abstract_annotations().items():
            if not _has_fn(impl, cls, name, expected):
                msg = (
                    f"{impl!r} does not fully implement {cls.__name__!r}: "
                    f"missing or mismatched {name!r}"
                )
                raise TypeError(msg)

    def _abstract_annotations(cls) -> dict[str, dict[str, object]]:
        methods: dict[str, dict[str, object]] = {}
        for base in cls.__mro__:
            if not isinstance(base, AnnotatedABCMeta):
                continue
            for name, value in base.__dict__.items():
                if getattr(value, "__isabstractmethod__", False):
                    methods[name] = _normalize_annotations(value)
        return methods


class AnnotatedABC(ABC, metaclass=AnnotatedABCMeta):
    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        for base in cls.__bases__:
            if isinstance(base, AnnotatedABCMeta):
                base._validate_implementation(cls)



from __future__ import annotations

import importlib
import sys

import pytest
from abc import abstractmethod

from abcreg import AnnotatedABC


class Queryable(AnnotatedABC):
    @abstractmethod
    def query(self, **kwargs) -> str: ...


class BaseQueryable(AnnotatedABC):
    @abstractmethod
    def query(self, **kwargs) -> str: ...


class ExtendedQueryable(BaseQueryable):
    @abstractmethod
    def query(self, **kwargs) -> str: ...

    @abstractmethod
    def fetch(self) -> int: ...


def test_direct_subclass_matching_annotations() -> None:
    class Good(Queryable):
        def query(self, **kwargs) -> str:
            return "ok"

    assert issubclass(Good, Queryable)
    assert Good().query() == "ok"


def test_direct_subclass_missing_method() -> None:
    with pytest.raises(TypeError, match="missing or mismatched 'query'"):

        class Bad(Queryable):
            pass


def test_direct_subclass_wrong_return_annotation() -> None:
    with pytest.raises(TypeError, match="missing or mismatched 'query'"):

        class Bad(Queryable):
            def query(self, **kwargs) -> int:
                return 1


def test_virtual_register_compliant() -> None:
    class Legacy:
        def query(self, **kwargs) -> str:
            return "legacy"

    Queryable.register(Legacy)
    assert issubclass(Legacy, Queryable)
    assert isinstance(Legacy(), Queryable)


def test_virtual_register_non_compliant() -> None:
    class Bad:
        def query(self, **kwargs) -> int:
            return 1

    with pytest.raises(TypeError, match="missing or mismatched 'query'"):
        Queryable.register(Bad)
    assert not issubclass(Bad, Queryable)


def test_inherited_abstract_methods() -> None:
    class Good(ExtendedQueryable):
        def query(self, **kwargs) -> str:
            return "ok"

        def fetch(self) -> int:
            return 42

    assert Good().fetch() == 42


def test_inherited_abstract_methods_missing() -> None:
    with pytest.raises(TypeError, match="missing or mismatched 'fetch'"):

        class Bad(ExtendedQueryable):
            def query(self, **kwargs) -> str:
                return "ok"


def test_abstract_intermediate_abc() -> None:
    class Partial(Queryable):
        @abstractmethod
        def query(self, **kwargs) -> str: ...

    assert issubclass(Partial, Queryable)


def test_get_annotations_from_inspect_on_python_before_3_14() -> None:
    if sys.version_info >= (3, 14):
        pytest.skip("get_annotations lives in annotationlib on Python 3.14+")

    import inspect

    from abcreg import _annotations

    importlib.reload(_annotations)
    assert _annotations.get_annotations is inspect.get_annotations


def test_get_annotations_from_annotationlib_on_python_3_14() -> None:
    if sys.version_info < (3, 14):
        pytest.skip("annotationlib.get_annotations is only available on Python 3.14+")

    import annotationlib

    from abcreg import _annotations

    importlib.reload(_annotations)
    assert _annotations.get_annotations is annotationlib.get_annotations
